@@ -570,6 +570,9 @@ class Agent:
         # Set up job context (support both job_id and prompt_id for compatibility)
         execution_id = task.get("prompt_id") or task.get("job_id") or str(uuid.uuid4())
         self.current_job_id = execution_id  # Keep internal name for now
+        
+        # Extract trace_id from task if present (for batch operations)
+        self.current_trace_id = task.get("trace_id")
 
         self.memory = AgentMemory(
             agent_name=self.config.name,
@@ -1610,6 +1613,10 @@ Provide a clear, complete final answer based on this information."""
         if event_type in self._trace_mask:
             return
         try:
+            # Include trace_id in payload if available (for batch operations)
+            if hasattr(self, 'current_trace_id') and self.current_trace_id:
+                payload = {**payload, "trace_id": self.current_trace_id}
+            
             self.database.append_trace(job_id, agent_name, event_type, payload)
         except Exception:
             # Tracing must never break runtime
