@@ -236,6 +236,11 @@ class LaddrConfig(BaseSettings):
         default=None, description="Dashboard user credentials (user:pass)"
     )
 
+    # API Key for protecting API routes (optional)
+    laddr_api_key: str | None = Field(
+        default=None, description="API key for protecting API routes. If set, all routes require this key."
+    )
+
     class Config:
         env_file = ".env"
         env_file_encoding = "utf-8"
