@@ -6,6 +6,13 @@ Provides tools for:
 - Storing large payloads in artifact registry (MinIO/S3)
 - Retrieving artifacts by ID
 - Tool override system for custom implementations
+
+Exported classes for user extensions:
+- TaskDelegationTool: Base class for single-task delegation
+- ParallelDelegationTool: Base class for parallel multi-task delegation
+- ArtifactStorageTool: Base class for artifact storage/retrieval
+
+Users can import these to build custom tool overrides that reuse base functionality.
 """
 
 from __future__ import annotations
@@ -19,6 +26,21 @@ import uuid
 # Module logger
 logger = logging.getLogger(__name__)
 
+# Public API exports
+__all__ = [
+    # Tool override system
+    "override_system_tool",
+    "get_tool_override",
+    "clear_tool_overrides",
+    "list_tool_overrides",
+    # Base tool classes for user extensions
+    "TaskDelegationTool",
+    "ParallelDelegationTool",
+    "ArtifactStorageTool",
+    # System tool factory
+    "create_system_tools",
+]
+
 # Global registry for tool overrides
 _TOOL_OVERRIDES: dict[str, Callable] = {}
 
@@ -36,7 +58,7 @@ def override_system_tool(tool_name: str):
         @override_system_tool("system_delegate_parallel")
         async def my_custom_parallel_delegation(agent_name: str, tasks: list[str], **kwargs):
             # Your custom implementation
-            print(f"Custom parallel delegation: {len(tasks)} tasks to {agent_name}")
+            logger.info(f"Custom parallel delegation: {len(tasks)} tasks to {agent_name}")
             return {"status": "success", "results": [...]}
     
     Args:
@@ -111,6 +133,12 @@ class TaskDelegationTool:
         self.artifact_storage = artifact_storage
         self.agent = agent
         self._size_threshold_kb = 100  # Store payloads > 100KB in artifact registry
+        
+        # Get storage bucket from agent config, fallback to default
+        self._storage_bucket = None
+        if agent and hasattr(agent, 'env_config'):
+            self._storage_bucket = getattr(agent.env_config, 'storage_bucket', None)
+        self._storage_bucket = self._storage_bucket or "laddr"
 
     async def delegate_task(
         self,
@@ -149,7 +177,6 @@ class TaskDelegationTool:
         - You do NOT need to call retrieve_artifact unless response explicitly says {"stored_in_registry": true}
         - task_id is for tracking only, NOT for artifact retrieval
         """
-        import sys
         # Backward compatibility: accept "timeout" as alias for "timeout_seconds"
         if timeout is not None:
             timeout_seconds = timeout
@@ -176,14 +203,19 @@ class TaskDelegationTool:
         # DEBUG: Show agent identity and state
         agent_id = id(self.agent) if self.agent else None
         agent_name_debug = getattr(self.agent.config, 'name', 'unknown') if self.agent and hasattr(self.agent, 'config') else 'unknown'
-        print(f"[DELEGATION DEBUG] agent_id={agent_id}, agent_name={agent_name_debug}, has_current_job_id={hasattr(self.agent, 'current_job_id') if self.agent else False}, agent.job_id={self.agent.current_job_id if self.agent and hasattr(self.agent, 'current_job_id') else None}, injected_parent_job_id={parent_job_id}, final_job_id={job_id_to_use}", file=sys.stderr, flush=True)
+        logger.debug(
+            f"Delegation: agent_id={agent_id}, agent_name={agent_name_debug}, "
+            f"has_current_job_id={hasattr(self.agent, 'current_job_id') if self.agent else False}, "
+            f"agent.job_id={self.agent.current_job_id if self.agent and hasattr(self.agent, 'current_job_id') else None}, "
+            f"injected_parent_job_id={parent_job_id}, final_job_id={job_id_to_use}"
+        )
         
         if job_id_to_use:
             if "job_id" not in payload:
                 payload["job_id"] = job_id_to_use
-                print(f"[DELEGATION] ✓ Propagating job_id={job_id_to_use} to {agent_name}", file=sys.stderr, flush=True)
+                logger.debug(f"Propagating job_id={job_id_to_use} to {agent_name}")
         else:
-            print(f"[DELEGATION WARNING] ✗ No job_id to propagate from agent {agent_name_debug}", file=sys.stderr, flush=True)
+            logger.warning(f"No job_id to propagate from agent {agent_name_debug}")
 
         # Check if payload is too large and should be stored in artifact registry
         payload_size_kb = len(json.dumps(payload).encode()) / 1024
@@ -253,8 +285,10 @@ class TaskDelegationTool:
         artifact_id = str(uuid.uuid4())
 
         if self.artifact_storage:
+            # Use configured bucket from env, not hardcoded
+            bucket = self._storage_bucket
             await self.artifact_storage.put_object(
-                bucket="laddr-artifacts",
+                bucket=bucket,
                 key=f"tasks/{artifact_id}.json",
                 data=json.dumps(data).encode()
             )
@@ -267,12 +301,15 @@ class TaskDelegationTool:
             return None
 
         try:
+            # Use configured bucket from env, not hardcoded
+            bucket = self._storage_bucket
             data = await self.artifact_storage.get_object(
-                bucket="laddr-artifacts",
+                bucket=bucket,
                 key=f"tasks/{artifact_id}.json"
             )
             return json.loads(data.decode())
-        except Exception:
+        except Exception as e:
+            logger.debug(f"Failed to retrieve artifact {artifact_id}: {e}")
             return None
 
 
@@ -284,7 +321,7 @@ class ArtifactStorageTool:
     - New style: by bucket+key (e.g., offloaded response pointers)
     """
 
-    def __init__(self, storage_backend):
+    def __init__(self, storage_backend, default_bucket: str | None = None):
         """
         Initialize artifact storage tool.
         
@@ -292,6 +329,10 @@ class ArtifactStorageTool:
             storage_backend: Storage backend (MinIO/S3)
         """
         self.storage = storage_backend
+        # Allow caller to provide a project-wide default bucket name
+        # (e.g. from LaddrConfig.storage_bucket). If not provided we
+        # keep backward-compatible defaults in retrieval logic.
+        self.default_bucket = default_bucket
 
     async def store_artifact(
         self,
@@ -320,8 +361,8 @@ class ArtifactStorageTool:
         else:
             data_bytes = data
 
-        # Store in artifact registry
-        bucket = "laddr-artifacts"
+        # Use configured bucket from env, fallback to legacy name
+        bucket = self.default_bucket or "laddr"
         key = f"{artifact_type}/{artifact_id}"
 
         await self.storage.put_object(
@@ -368,11 +409,10 @@ class ArtifactStorageTool:
             Dict with artifact data, metadata, and retrieval info
         """
         # New path: direct bucket/key retrieval (e.g., offloaded responses)
-        # NOTE: artifact storage uses the bucket "laddr-artifacts" by default
-        # for stored task artifacts. Use that as the default to avoid confusing
-        # callers that pass only a key or a job id.
+        # Prefer an explicit bucket argument, then the configured default
+        # bucket (self.default_bucket), then fall back to legacy names.
         if key is not None:
-            use_bucket = bucket or "laddr-artifacts"
+            use_bucket = bucket or self.default_bucket or "laddr"
             try:
                 data = await self.storage.get_object(bucket=use_bucket, key=key)
                 try:
@@ -388,19 +428,32 @@ class ArtifactStorageTool:
                     "retrieved_at": datetime.utcnow().isoformat(),
                 }
             except Exception as e:
-                return {
-                    "bucket": use_bucket,
-                    "key": key,
-                    "error": str(e),
-                    "status": "not_found",
-                }
+                # Try common legacy fallback bucket if first attempt fails
+                fallback_bucket = "laddr-artifacts"
+                try:
+                    data = await self.storage.get_object(bucket=fallback_bucket, key=key)
+                    try:
+                        parsed_data = json.loads(data.decode())
+                    except (json.JSONDecodeError, UnicodeDecodeError):
+                        parsed_data = data.decode(errors='replace')
+
+                    return {
+                        "bucket": fallback_bucket,
+                        "key": key,
+                        "data": parsed_data,
+                        "size_bytes": len(data),
+                        "retrieved_at": datetime.utcnow().isoformat(),
+                    }
+                except Exception:
+                    # Not found anywhartifactsere -> raise so caller/runtime treats this as a tool error
+                    raise FileNotFoundError(f"Artifact not found at {use_bucket}/{key} (checked fallback {fallback_bucket})")
 
         # Detect if artifact_id is actually a full key path (contains slashes)
         # This handles the case where an offloaded response key is passed as artifact_id
         if artifact_id and "/" in artifact_id:
             # This is likely a response key like "responses/2025/11/04/<task_id>.json"
-            # Response storage uses the main bucket (not laddr-artifacts)
-            use_bucket = bucket or "laddr"  # Default to main bucket for responses
+            # Prefer explicit bucket, then configured default bucket, then legacy main bucket name
+            use_bucket = bucket or self.default_bucket or "laddr"
             try:
                 data = await self.storage.get_object(bucket=use_bucket, key=artifact_id)
                 try:
@@ -416,7 +469,7 @@ class ArtifactStorageTool:
                     "retrieved_at": datetime.utcnow().isoformat(),
                 }
             except Exception as e:
-                # If not found in main bucket, try laddr-artifacts as fallback
+                # If not found in preferred bucket, try legacy artifacts bucket
                 try:
                     fallback_bucket = "laddr-artifacts"
                     data = await self.storage.get_object(bucket=fallback_bucket, key=artifact_id)
@@ -433,25 +486,19 @@ class ArtifactStorageTool:
                         "retrieved_at": datetime.utcnow().isoformat(),
                     }
                 except Exception:
-                    # Return original error
-                    return {
-                        "bucket": use_bucket,
-                        "key": artifact_id,
-                        "error": str(e),
-                        "status": "not_found",
-                    }
+                    # Not found in any expected location -> raise for runtime to record a tool error
+                    raise FileNotFoundError(f"Artifact not found at {use_bucket}/{artifact_id} (checked fallback laddr-artifacts)")
 
         # Legacy path: artifact registry lookup by id/type
-        bucket = "laddr-artifacts"
+        # Use configured bucket from env, fallback to legacy name
+        legacy_bucket = self.default_bucket or "laddr"
         if not artifact_id:
-            return {
-                "error": "artifact_id or (bucket and key) is required",
-                "status": "invalid_params",
-            }
+            raise ValueError("artifact_id or (bucket and key) is required")
+
         legacy_key = f"{artifact_type}/{artifact_id}"
 
         try:
-            data = await self.storage.get_object(bucket=bucket, key=legacy_key)
+            data = await self.storage.get_object(bucket=legacy_bucket, key=legacy_key)
             try:
                 parsed_data = json.loads(data.decode())
             except (json.JSONDecodeError, UnicodeDecodeError):
@@ -465,11 +512,8 @@ class ArtifactStorageTool:
                 "retrieved_at": datetime.utcnow().isoformat(),
             }
         except Exception as e:
-            return {
-                "artifact_id": artifact_id,
-                "error": str(e),
-                "status": "not_found",
-            }
+            # Not found -> raise so caller/runtime treats as tool error
+            raise FileNotFoundError(f"Legacy artifact not found at {legacy_bucket}/{legacy_key}: {e}")
 
     async def list_artifacts(
         self,
@@ -486,7 +530,8 @@ class ArtifactStorageTool:
         Returns:
             Dict with list of artifact metadata
         """
-        bucket = "laddr-artifacts"
+        # Use configured bucket from env, fallback to legacy name
+        bucket = self.default_bucket or "laddr"
         prefix = f"{artifact_type}/" if artifact_type else ""
 
         try:
@@ -750,7 +795,7 @@ def create_system_tools(message_bus, storage_backend=None, agent=None) -> dict[s
     
     This function checks for user-provided overrides before registering
     the default implementations. Users can use @override_system_tool 
-    decorator to replace any system tool.
+    decorator to replace any system tool OR create entirely new system tools.
     
     Args:
         message_bus: Message queue backend
@@ -760,13 +805,51 @@ def create_system_tools(message_bus, storage_backend=None, agent=None) -> dict[s
     Returns:
         Dict mapping tool names to (tool_function, aliases) tuples
     """
+    import functools
+    import inspect as _inspect
+    
+    def wrap_with_runtime_injection(tool_func):
+        """
+        Wrap a custom system tool function to inject runtime parameters.
+        
+        Custom override functions can declare _message_bus, _artifact_storage, _agent
+        parameters which will be automatically injected at call time.
+        """
+        sig = _inspect.signature(tool_func)
+        needs_injection = any(p in sig.parameters for p in ['_message_bus', '_artifact_storage', '_agent'])
+        
+        if not needs_injection:
+            # No injection needed, return as-is
+            return tool_func
+        
+        # Create wrapper that injects runtime parameters
+        @functools.wraps(tool_func)
+        async def wrapped_tool(*args, **kwargs):
+            # Inject runtime parameters if not already provided
+            if '_message_bus' in sig.parameters and '_message_bus' not in kwargs:
+                kwargs['_message_bus'] = message_bus
+            if '_artifact_storage' in sig.parameters and '_artifact_storage' not in kwargs:
+                kwargs['_artifact_storage'] = storage_backend
+            if '_agent' in sig.parameters and '_agent' not in kwargs:
+                kwargs['_agent'] = agent
+            
+            # Call the original function
+            result = tool_func(*args, **kwargs)
+            
+            # Await if async
+            if _inspect.iscoroutine(result) or _inspect.isawaitable(result):
+                return await result
+            return result
+        
+        return wrapped_tool
+    
     tools = {}
 
     # Task delegation tool (with agent context for job_id propagation)
     override = get_tool_override("system_delegate_task")
     if override:
         logger.info("✅ Using custom override for system_delegate_task")
-        tools["system_delegate_task"] = (override, [])
+        tools["system_delegate_task"] = (wrap_with_runtime_injection(override), [])
     else:
         delegation_tool = TaskDelegationTool(message_bus, storage_backend, agent=agent)
         tools["system_delegate_task"] = (delegation_tool.delegate_task, [])
@@ -778,7 +861,7 @@ def create_system_tools(message_bus, storage_backend=None, agent=None) -> dict[s
     override = get_tool_override("system_delegate_parallel")
     if override:
         logger.info("✅ Using custom override for system_delegate_parallel")
-        tools["system_delegate_parallel"] = (override, ["delegate_parallel"])
+        tools["system_delegate_parallel"] = (wrap_with_runtime_injection(override), ["delegate_parallel"])
     else:
         tools["system_delegate_parallel"] = (
             parallel_tool.delegate_parallel,
@@ -788,7 +871,7 @@ def create_system_tools(message_bus, storage_backend=None, agent=None) -> dict[s
     override = get_tool_override("system_split_document")
     if override:
         logger.info("✅ Using custom override for system_split_document")
-        tools["system_split_document"] = (override, ["split_document"])
+        tools["system_split_document"] = (wrap_with_runtime_injection(override), ["split_document"])
     else:
         tools["system_split_document"] = (
             parallel_tool.split_document,
@@ -797,27 +880,47 @@ def create_system_tools(message_bus, storage_backend=None, agent=None) -> dict[s
 
     # Artifact storage tools (if storage backend available)
     if storage_backend:
-        artifact_tool = ArtifactStorageTool(storage_backend)
+        # Pass configured default bucket when available (agent may be None)
+        default_bucket = None
+        try:
+            default_bucket = agent.env_config.storage_bucket if agent and getattr(agent, 'env_config', None) else None
+        except Exception:
+            default_bucket = None
+        artifact_tool = ArtifactStorageTool(storage_backend, default_bucket=default_bucket)
         
         override = get_tool_override("system_store_artifact")
         if override:
             logger.info("✅ Using custom override for system_store_artifact")
-            tools["system_store_artifact"] = (override, [])
+            tools["system_store_artifact"] = (wrap_with_runtime_injection(override), [])
         else:
             tools["system_store_artifact"] = (artifact_tool.store_artifact, [])
         
         override = get_tool_override("system_retrieve_artifact")
         if override:
             logger.info("✅ Using custom override for system_retrieve_artifact")
-            tools["system_retrieve_artifact"] = (override, [])
+            tools["system_retrieve_artifact"] = (wrap_with_runtime_injection(override), [])
         else:
             tools["system_retrieve_artifact"] = (artifact_tool.retrieve_artifact, [])
         
         override = get_tool_override("system_list_artifacts")
         if override:
             logger.info("✅ Using custom override for system_list_artifacts")
-            tools["system_list_artifacts"] = (override, [])
+            tools["system_list_artifacts"] = (wrap_with_runtime_injection(override), [])
         else:
             tools["system_list_artifacts"] = (artifact_tool.list_artifacts, [])
 
+    # NEW FEATURE: Register any additional custom system tools that weren't overrides
+    # This allows users to create entirely new system tools, not just override existing ones
+    known_tool_names = set(tools.keys())
+    for tool_name, tool_func in _TOOL_OVERRIDES.items():
+        # Skip aliases (like "delegate_parallel") - they're registered with the main tool
+        if tool_name in ["delegate_parallel", "split_document", "delegate_task"]:
+            continue
+        
+        # Add new custom system tools that don't override built-in ones
+        if tool_name not in known_tool_names:
+            logger.info(f"✅ Registering NEW system tool: {tool_name}")
+            # Wrap with runtime injection
+            tools[tool_name] = (wrap_with_runtime_injection(tool_func), [])
+
     return tools
